# `encrustable` - basic components of Rust's type system magic
# Copyright (C) 2025 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import NewType

from encrustable import Nothing, Option, Some

# we use the newtype pattern to squeeze the most ouf of our type checker
Username = NewType("Username", str)
PasswdFileLine = NewType("PasswdFileLine", str)
ParsedPasswdEntry = NewType("ParsedPasswdEntry", list[str])
ShellPath = NewType("ShellPath", str)


# we assume the passwd file exists, is readable and consists of valid lines, so we have
# a function that can fail in exactly one way (the user is not in the passwd file)
# we encode this scenario as an Option
def get_user_entry(username: Username) -> Option[PasswdFileLine]:
    with open("/etc/passwd", "r") as passwd_file:
        for line in passwd_file:
            parts: list[str] = line.split(":")
            if parts[0] == username:
                return Some(PasswdFileLine(line.strip()))
        return Nothing()


# we have a regular function that cannot fail
# assuming it receives a valid passwd file line
def parse_user_entry(entry: PasswdFileLine) -> ParsedPasswdEntry:
    return ParsedPasswdEntry(entry.split(":"))


# we have another regular function that cannot fail
# assuming it receives a valid parsed passwd entry
def get_shell_name(parsed_entry: ParsedPasswdEntry) -> ShellPath:
    return ShellPath(parsed_entry[6])


# now we can combine them and we can skip any error checking midway
# we will only check the result at the end using pattern matching
def print_user_shell(username: Username) -> None:
    match get_user_entry(username) | parse_user_entry | get_shell_name:
        # the result is of the type `Option[ShellPath]`
        case Some(s):
            print(f"- user '{username}' has shell set to '{s}'")
        case Nothing():
            print(f"- user '{username}' not found in the '/etc/passwd' file")


# execute the example for two users, "root" and "anchovies"
if __name__ == "__main__":
    print_user_shell(Username("root"))
    print_user_shell(Username("anchovies"))



# `encrustable` - basic components of Rust's type system magic
# Copyright (C) 2025 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from collections.abc import Callable, Generator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Never, Protocol, final

from encrustable.panic import Panic

if TYPE_CHECKING:  # pragma: nocover
    from encrustable.option import Option


class ResultProto[T, E: Exception](Protocol):
    """
    This is a common protocol for the `Ok` and `Err` variants that defines
    all available operations for `Result`s.
    """

    def __iter__(self) -> Generator[T]:  # pragma: nocover
        """
        Returns a one-time-use iterator object that will yield
        a single value if `Ok`, otherwise no values.
        """
        ...

    def __or__[V](self, fn: Callable[[T], V]) -> Result[V, E]:
        """
        This implements pipe operator `|` as a convenient alias to the `.map` method.
        """
        return self.map(fn)

    def is_ok(self) -> bool:  # pragma: nocover
        """
        Returns `True` if `Ok`, otherwise `False`.
        """
        ...

    def is_ok_and(self, p: Callable[[T], bool]) -> bool:  # pragma: nocover
        """
        Returns `True` if `Ok` and the predicate `p` returns `True` when applied to
        the contained value, otherwise `False`.
        """
        ...

    def is_err(self) -> bool:  # pragma: nocover
        """
        Returns `True` if `Err`, otherwise `False`.
        """
        ...

    def is_err_and(self, p: Callable[[E], bool]) -> bool:  # pragma: nocover
        """
        Returns `True` if `Err` and the predicate `p` returns `True` when applied to
        the contained error, otherwise `False`.
        """
        ...

    def and_[V](self, r: Result[V, E]) -> Result[V, E]:  # pragma: nocover
        """
        Returns the `Result` in `r` if `Ok`, otherwise the `Err` variant with the
        contained error.

        Note: `and` is a reserved Python keyword, so the method is called `and_`.
        """
        ...

    def and_then[V](
        self, f: Callable[[], Result[V, E]]
    ) -> Result[V, E]:  # pragma: nocover
        """
        Returns the result of calling the factory in `f` if `Ok`, otherwise the
        `Err` variant with the contained error.
        """
        ...

    def or_[F: Exception](self, r: Result[T, F]) -> Result[T, F]:  # pragma: nocover
        """
        Returns `self` if `Some`, otherwise the `Option` in `o`.

        Note: `or` is a reserved Python keyword, so the method is called `or_`.
        """
        ...

    def or_else[F: Exception](
        self, f: Callable[[], Result[T, F]]
    ) -> Result[T, F]:  # pragma: nocover
        """
        Returns `self` if `Some`, otherwise the result of calling the factory in `f`.
        """
        ...

    def inspect(self, fn: Callable[[T], None]) -> Result[T, E]:  # pragma: nocover
        """
        Returns `self` after applying the function `fn` to the contained value if `Ok`.
        """
        ...

    def inspect_err(self, fn: Callable[[E], None]) -> Result[T, E]:  # pragma: nocover
        """
        Returns `self` after applying the function `fn` to the contained error if `Err`.
        """
        ...

    def unwrap(self) -> T:  # pragma: nocover
        """
        Returns the contained value if `Ok`, otherwise raises a `Panic`.
        """
        ...

    def unwrap_err(self) -> E:  # pragma: nocover
        """
        Returns the contained error if `Err`, otherwise raises a `Panic`.
        """
        ...

    def unwrap_or(self, t: T) -> T:  # pragma: nocover
        """
        Returns the contained value if `Ok`, otherwise the value `t`.
        """
        ...

    def unwrap_or_else(self, f: Callable[[], T]) -> T:  # pragma: nocover
        """
        Returns the contained value if `Ok`, otherwise the result of
        calling the factory in `f`.
        """
        ...

    def expect(self, m: str) -> T:  # pragma: nocover
        """
        Returns the contained value if `Ok`, otherwise raises a `Panic`
        with a custom message.
        """
        ...

    def expect_err(self, m: str) -> E:  # pragma: nocover
        """
        Returns the contained error if `Err`, otherwise raises a `Panic`
        with a custom message.
        """
        ...

    def map[V](self, fn: Callable[[T], V]) -> Result[V, E]:  # pragma: nocover
        """
        Returns a new `Result` containing the result of application of the function in
        `fn` to the contained value if `Ok`, otherwise the `Result` untouched.
        """
        ...

    def map_err[F: Exception](
        self, fn: Callable[[E], F]
    ) -> Result[T, F]:  # pragma: nocover
        """
        Returns a new `Result` containing the result of application of the function in
        `fn` to the contained error if `Err`, otherwise the `Result` untouched.
        """
        ...

    def map_or[V](self, v: V, fn: Callable[[T], V]) -> V:  # pragma: nocover
        """
        Returns the result of application of the function in `fn` to the contained
        value if `Ok`, otherwise the value in `v`.
        """
        ...

    def map_or_else[V](
        self, v: Callable[[], V], fn: Callable[[T], V]
    ) -> V:  # pragma: nocover
        """
        Returns the result of application of the function in `fn` to the contained
        value if `Ok`, otherwise the result of calling the factory in `v`.
        """
        ...

    def ok(self) -> Option[T]:  # pragma: nocover
        """
        Converts this `Result` to an `Option[T]` - returns `Some` with the contained
        value if `Ok`, otherwise `Nothing` (contained error is discarded).
        """
        ...

    def err(self) -> Option[E]:  # pragma: nocover
        """
        Converts this `Result` to an `Option[E]` - returns `Some` with the contained
        error if `Err`, otherwise `Nothing`.
        """
        ...

    def flatten[V, F: Exception](self: Result[Result[V, F], F]) -> Result[V, F]:
        """
        Converts a `Result[Result[T, E], E]` to a flat `Result[T, E]`.
        """
        match self:
            case Ok(r):
                match r:
                    case Ok(v):
                        return Ok(v)
                    case Err(e):  # pragma: nobranch
                        return Err(e)
            case Err(e):  # pragma: nobranch
                return Err(e)

    def transpose[V, F: Exception](self: Result[Option[V], F]) -> Option[Result[V, F]]:
        """
        Converts a `Result[Option[T], E]` to an `Option[Result[T, E]]`.
        """
        from encrustable.option import Nothing, Some

        match self:
            case Ok(o):
                match o:
                    case Some(v):
                        return Some(Ok(v))
                    case Nothing():  # pragma: nobranch
                        return Nothing()
            case Err(e):  # pragma: nobranch
                return Some(Err(e))


@final
@dataclass(frozen=True, eq=True)
class Ok[T, E: Exception](ResultProto[T, E]):
    """
    This represents the variant that holds a result of a successful computation.
    """

    v: T

    def __iter__(self) -> Generator[T]:
        yield self.v

    def is_ok(self) -> bool:
        return True

    def is_ok_and(self, p: Callable[[T], bool]) -> bool:
        return p(self.v)

    def is_err(self) -> bool:
        return False

    def is_err_and(self, p: Callable[[E], bool]) -> bool:
        return False

    def and_[V](self, r: Result[V, E]) -> Result[V, E]:
        return r

    def and_then[V](self, f: Callable[[], Result[V, E]]) -> Result[V, E]:
        return f()

    def or_[F: Exception](self, r: Result[T, F]) -> Result[T, F]:
        return Ok(self.v)

    def or_else[F: Exception](self, f: Callable[[], Result[T, F]]) -> Result[T, F]:
        return Ok(self.v)

    def inspect(self, fn: Callable[[T], None]) -> Result[T, E]:
        fn(self.v)
        return self

    def inspect_err(self, fn: Callable[[E], None]) -> Result[T, E]:
        return self

    def unwrap(self) -> T:
        return self.v

    def unwrap_err(self) -> Never:
        raise Panic(str(self))

    def unwrap_or(self, t: T) -> T:
        return self.v

    def unwrap_or_else(self, f: Callable[[], T]) -> T:
        return self.v

    def expect(self, m: str) -> T:
        return self.v

    def expect_err(self, m: str) -> Never:
        raise Panic(m)

    def map[V](self, fn: Callable[[T], V]) -> Result[V, E]:
        return Ok(fn(self.v))

    def map_err[F: Exception](self, fn: Callable[[E], F]) -> Result[T, F]:
        return Ok(self.v)

    def map_or[V](self, v: V, fn: Callable[[T], V]) -> V:
        return fn(self.v)

    def map_or_else[V](self, v: Callable[[], V], fn: Callable[[T], V]) -> V:
        return fn(self.v)

    def ok(self) -> Option[T]:
        from encrustable.option import Some

        return Some(self.v)

    def err(self) -> Option[E]:
        from encrustable.option import Nothing

        return Nothing()


@final
@dataclass(frozen=True, eq=True)
class Err[T, E: Exception](ResultProto[T, E]):
    """
    This represents the variant that holds an error that occured during a computation.
    """

    e: E

    def __iter__(self) -> Generator[T]:
        return
        yield

    def __or__[V](self, fn: Callable[[T], V]) -> Result[V, E]:
        return Err(self.e)

    def is_ok(self) -> bool:
        return False

    def is_ok_and(self, p: Callable[[T], bool]) -> bool:
        return False

    def is_err(self) -> bool:
        return True

    def is_err_and(self, p: Callable[[E], bool]) -> bool:
        return p(self.e)

    def and_[V](self, r: Result[V, E]) -> Result[V, E]:
        return Err(self.e)

    def and_then[V](self, f: Callable[[], Result[V, E]]) -> Result[V, E]:
        return Err(self.e)

    def or_[F: Exception](self, r: Result[T, F]) -> Result[T, F]:
        return r

    def or_else[F: Exception](self, f: Callable[[], Result[T, F]]) -> Result[T, F]:
        return f()

    def inspect(self, fn: Callable[[T], None]) -> Result[T, E]:
        return self

    def inspect_err(self, fn: Callable[[E], None]) -> Result[T, E]:
        fn(self.e)
        return self

    def unwrap(self) -> Never:
        raise Panic() from self.e

    def unwrap_err(self) -> E:
        return self.e

    def unwrap_or(self, t: T) -> T:
        return t

    def unwrap_or_else(self, f: Callable[[], T]) -> T:
        return f()

    def expect(self, m: str) -> Never:
        raise Panic(m) from self.e

    def expect_err(self, m: str) -> E:
        return self.e

    def map[V](self, fn: Callable[[T], V]) -> Result[V, E]:
        return Err(self.e)

    def map_err[F: Exception](self, fn: Callable[[E], F]) -> Result[T, F]:
        return Err(fn(self.e))

    def map_or[V](self, v: V, fn: Callable[[T], V]) -> V:
        return v

    def map_or_else[V](self, v: Callable[[], V], fn: Callable[[T], V]) -> V:
        return v()

    def ok(self) -> Option[T]:
        from encrustable.option import Nothing

        return Nothing()

    def err(self) -> Option[E]:
        from encrustable.option import Some

        return Some(self.e)


type Result[T, E: Exception] = Ok[T, E] | Err[T, E]



# `encrustable` - basic components of Rust's type system magic
# Copyright (C) 2025 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import final


@final
@dataclass(frozen=True, eq=True)
class Panic(Exception):
    """
    A `Panic` is an error raised when the type system assumptions are
    somehow violated. It's a standard Python `Exception` and can be caught
    at the boundary. The error message (custom if provided, or default)
    can be extracted by casting the error to `str`.
    """

    message: str | None = None
    """A custom panic message, if provided."""

    def __str__(self) -> str:
        return self.message or "panic!"



# `encrustable` - basic components of Rust's type system magic
# Copyright (C) 2025 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import tempfile
from dataclasses import dataclass, field
from pathlib import Path

from encrustable import Err, Ok, Result


@dataclass(frozen=True, eq=True)
class FileReadError(Exception):
    file_name: Path
    original_error: Exception | None = field(default=None, kw_only=True, repr=False)


class FileNotFound(FileReadError):
    pass


class FileNotReadable(FileReadError):
    pass


class FileEmpty(FileReadError):
    pass


# this function can fail in many different ways:
#   - file does not exist
#   - file is not readable (because of permissions)
#   - file is empty (so no first line)
#   - we can even have a catch-all generic exception handler if we want to
# we encode this scenario as a Result
def read_first_line_from_file(file_name: Path) -> Result[str, FileReadError]:
    try:
        with file_name.open("r") as file:
            if (line := file.readline().strip()) == "":
                return Err(FileEmpty(file_name))
            return Ok(line)
    except FileNotFoundError as e:
        return Err(FileNotFound(file_name, original_error=e))
    except PermissionError as e:
        return Err(FileNotReadable(file_name, original_error=e))
    except Exception as e:
        return Err(FileReadError(file_name, original_error=e))


# we have a regular function that cannot fail
# given a valid string it will give us up to 20 characters from the beginning
def trim_str_above_20_len(line: str) -> str:
    return line.strip()[:20]


# now we can combine them and we can skip any error checking midway
# we will only check the result at the end using pattern matching
def print_first_line_from_file(file_name: Path) -> None:
    match read_first_line_from_file(file_name) | trim_str_above_20_len:
        case Ok(line):
            print(f"- First line (up to 20 chars) from '{file_name}': {line}")
        case Err(err):
            print(
                f"- Could not read first line from '{str(file_name)[:20]}'\n"
                f"  error: {repr(err)}\n"
                f"  original error: {repr(err.original_error)}"
            )


# execute the example for a couple of filenames
if __name__ == "__main__":
    print_first_line_from_file(Path("/etc/passwd"))
    print_first_line_from_file(Path("/does-not-exist.yaml"))
    print_first_line_from_file(Path("/root/file.yaml"))
    with tempfile.NamedTemporaryFile("w") as f:
        print_first_line_from_file(Path(f.name))



# `encrustable` - basic components of Rust's type system magic
# Copyright (C) 2025 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from encrustable import Err, Nothing, Ok, Option, Panic, Result, Some


def get_some(v: int) -> Option[int]:
    return Some(v)


def get_nothing() -> Option[int]:
    return Nothing()


def get_ok(v: int) -> Result[int, Exception]:
    return Ok(v)


def get_err(e: Exception) -> Result[int, Exception]:
    return Err(e)


def try_unwrap[T, E: Exception](v: Option[T] | Result[T, E]) -> None:
    try:
        unwrapped = v.unwrap()
        print(f"- {repr(v)} unwrap: {repr(unwrapped)}")
    except Panic as p:
        print(f"- {repr(v)} unwrap: {str(p)}\n  cause: {repr(p.__cause__)}")


def try_expect[T, E: Exception](v: Option[T] | Result[T, E]) -> None:
    try:
        unwrapped = v.expect("no valid value")
        print(f"- {repr(v)} expect: {repr(unwrapped)}")
    except Panic as p:
        print(f"- {repr(v)} expect: {str(p)}\n  cause: {repr(p.__cause__)}")


if __name__ == "__main__":
    try_unwrap(get_some(1))
    try_expect(get_some(1))

    try_unwrap(get_nothing())
    try_expect(get_nothing())

    try_unwrap(get_ok(1))
    try_expect(get_ok(1))

    try_unwrap(get_err(Exception()))
    try_expect(get_err(Exception()))



# `encrustable` - basic components of Rust's type system magic
# Copyright (C) 2025 Artur Ciesielski <artur.ciesielski@gmail.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from collections.abc import Callable, Generator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Never, Protocol, final

from encrustable.panic import Panic

if TYPE_CHECKING:  # pragma: nocover
    from encrustable.result import Result


class OptionProto[T](Protocol):
    """
    This is a protocol base for the `Some` and `Nothing` variants that defines
    all available operations for `Option`s.
    """

    def __iter__(self) -> Generator[T]:  # pragma: nocover
        """
        Returns a one-time-use iterator object that will yield
        a single value if `Some`, otherwise no values.
        """
        ...

    def __or__[V](self, fn: Callable[[T], V]) -> Option[V]:
        """
        This implements pipe operator `|` as a convenient alias to the `.map` method.
        """
        return self.map(fn)

    def is_some(self) -> bool:  # pragma: nocover
        """
        Returns `True` if `Some`, otherwise `False`.
        """
        ...

    def is_some_and(self, p: Callable[[T], bool]) -> bool:  # pragma: nocover
        """
        Returns `True` if `Some` and the predicate `p` returns `True` when applied to
        the contained value, otherwise `False`.
        """
        ...

    def is_nothing(self) -> bool:  # pragma: nocover
        """
        Returns `True` if `Nothing`, otherwise `False`.
        """
        ...

    def is_nothing_or(self, p: Callable[[T], bool]) -> bool:  # pragma: nocover
        """
        Returns `True` if `Nothing` or the predicate `p` returns `True` when applied to
        the contained value, otherwise `False`.
        """
        ...

    def and_[V](self, o: Option[V]) -> Option[V]:  # pragma: nocover
        """
        Returns the `Option` in `o` if `Some`, otherwise `Nothing`.

        Note: `and` is a reserved Python keyword, so the method is called `and_`.
        """
        ...

    def and_then[V](self, f: Callable[[], Option[V]]) -> Option[V]:  # pragma: nocover
        """
        Returns the result of calling the factory in `f` if `Some`, otherwise `Nothing`.
        """
        ...

    def or_(self, o: Option[T]) -> Option[T]:  # pragma: nocover
        """
        Returns `self` if `Some`, otherwise the `Option` in `o`.

        Note: `or` is a reserved Python keyword, so the method is called `or_`.
        """
        ...

    def or_else(self, f: Callable[[], Option[T]]) -> Option[T]:  # pragma: nocover
        """
        Returns `self` if `Some`, otherwise the result of calling the factory in `f`.
        """
        ...

    def or_none(self) -> T | None:  # pragma: nocover
        """
        Returns the contained value if `Some`, otherwise `None`.
        """
        ...

    def xor(self, o: Option[T]) -> Option[T]:  # pragma: nocover
        """
        Returns the `Some` variant with the contained value if exactly one of `self`
        and `o` is `Some`, otherwise `Nothing`.
        """
        ...

    def filter(self, p: Callable[[T], bool]) -> Option[T]:  # pragma: nocover
        """
        Returns `self` if `Some` and the predicate `p` returns `True` when applied to
        the contained value, otherwise `Nothing`.
        """
        ...

    def inspect(self, fn: Callable[[T], None]) -> Option[T]:  # pragma: nocover
        """
        Returns `self` after applying the function `fn` to the contained value
        if `Some`.
        """
        ...

    def unwrap(self) -> T:  # pragma: nocover
        """
        Returns the contained value if `Some`, otherwise raises a `Panic`.
        """
        ...

    def unwrap_or(self, t: T) -> T:  # pragma: nocover
        """
        Returns the contained value if `Some`, otherwise the value `t`.
        """
        ...

    def unwrap_or_else(self, f: Callable[[], T]) -> T:  # pragma: nocover
        """
        Returns the contained value if `Some`, otherwise the result of
        calling the factory in `f`.
        """
        ...

    def expect(self, m: str) -> T:  # pragma: nocover
        """
        Returns the contained value if `Some`, otherwise raises a `Panic`
        with a custom message.
        """
        ...

    def map[V](self, fn: Callable[[T], V]) -> Option[V]:  # pragma: nocover
        """
        Returns a new `Result` containing the result of application of the function in
        `fn` to the contained value if `Some`, otherwise `Nothing`.
        """
        ...

    def map_or[V](self, v: V, fn: Callable[[T], V]) -> V:  # pragma: nocover
        """
        Returns the result of application of the function in `fn` to the contained
        value if `Some`, otherwise the value in `v`.
        """
        ...

    def map_or_else[V](
        self, v: Callable[[], V], fn: Callable[[T], V]
    ) -> V:  # pragma: nocover
        """
        Returns the result of application of the function in `fn` to the contained
        value if `Some`, otherwise the result of calling the factory in `v`.
        """
        ...

    def ok_or[E: Exception](self, e: E) -> Result[T, E]:  # pragma: nocover
        """
        Converts this `Option` to a `Result[T, E]` - returns an `Ok` with the contained
        value if `Some`, otherwise an `Err` with the error `e`.
        """
        ...

    def ok_or_else[E: Exception](
        self, f: Callable[[], E]
    ) -> Result[T, E]:  # pragma: nocover
        """
        Converts this `Option` to a `Result[T, E]` - returns an `Ok` with the contained
        value if `Some`, otherwise an `Err` containing the result of calling the
        factory in `f`.
        """
        ...

    def zip[V](self, o: Option[V]) -> Option[tuple[T, V]]:  # pragma: nocover
        """
        Returns an `Option[(T, V)]` resulting from zipping contained values in this
        `Option` together with another `Option[V]` into a pair (2-tuple) of values.

        Returns `Nothing` if either `self` or `o` is `Nothing`.
        """
        ...

    def zip_with[V, R](self, o: Option[V], f: Callable[[T, V], R]) -> Option[R]:
        """
        Returns an `Option` of a value of type `R` resulting from zipping this option
        together with another `Option[V]` and applying a two-argument function `f`
        on the resulting pair of values.

        Returns `Nothing` if either `self` or `o` is `Nothing`.
        """
        return self.zip(o).map(lambda pair: f(pair[0], pair[1]))

    def flatten[O](self: Option[Option[O]]) -> Option[O]:
        """
        Converts an `Option[Option[T]]` to a flat `Option[T]`.
        """
        match self:
            case Some(o):
                match o:
                    case Some(v):
                        return Some(v)
                    case Nothing():  # pragma: nobranch
                        return Nothing()
            case Nothing():  # pragma: nobranch
                return Nothing()

    def transpose[V, F: Exception](self: Option[Result[V, F]]) -> Result[Option[V], F]:
        """
        Converts an `Option[Result[T, E]]` to a `Result[Option[T], E]`.
        """
        from encrustable.result import Err, Ok

        match self:
            case Some(r):
                match r:
                    case Ok(v):
                        return Ok(Some(v))
                    case Err(e):  # pragma: nobranch
                        return Err(e)
            case Nothing():  # pragma: nobranch
                return Ok(Nothing())


@final
@dataclass(frozen=True, eq=True)
class Some[T](OptionProto[T]):
    """
    This represents the variant that holds a value.
    """

    v: T

    def __iter__(self) -> Generator[T]:
        yield self.v

    def is_some(self) -> bool:
        return True

    def is_some_and(self, p: Callable[[T], bool]) -> bool:
        return p(self.v)

    def is_nothing(self) -> bool:
        return False

    def is_nothing_or(self, p: Callable[[T], bool]) -> bool:
        return p(self.v)

    def and_[V](self, o: Option[V]) -> Option[V]:
        return o

    def and_then[V](self, f: Callable[[], Option[V]]) -> Option[V]:
        return f()

    def or_(self, o: Option[T]) -> Option[T]:
        return self

    def or_else(self, f: Callable[[], Option[T]]) -> Option[T]:
        return self

    def or_none(self) -> T | None:
        return self.v

    def xor(self, o: Option[T]) -> Option[T]:
        return self if o.is_nothing() else Nothing()

    def filter(self, p: Callable[[T], bool]) -> Option[T]:
        return self if p(self.v) else Nothing()

    def inspect(self, fn: Callable[[T], None]) -> Option[T]:
        fn(self.v)
        return self

    def unwrap(self) -> T:
        return self.v

    def unwrap_or(self, t: T) -> T:
        return self.v

    def unwrap_or_else(self, f: Callable[[], T]) -> T:
        return self.v

    def expect(self, m: str) -> T:
        return self.v

    def map[V](self, fn: Callable[[T], V]) -> Option[V]:
        return Some(fn(self.v))

    def map_or[V](self, v: V, fn: Callable[[T], V]) -> V:
        return fn(self.v)

    def map_or_else[V](self, v: Callable[[], V], fn: Callable[[T], V]) -> V:
        return fn(self.v)

    def ok_or[E: Exception](self, e: E) -> Result[T, E]:
        from encrustable.result import Ok

        return Ok(self.v)

    def ok_or_else[E: Exception](self, f: Callable[[], E]) -> Result[T, E]:
        from encrustable.result import Ok

        return Ok(self.v)

    def zip[V](self, o: Option[V]) -> Option[tuple[T, V]]:
        return Some((self.v, o.unwrap())) if o.is_some() else Nothing()


@final
@dataclass(frozen=True, eq=True)
class Nothing[T](OptionProto[T]):
    """
    This represents the empty variant.
    """

    def __iter__(self) -> Generator[T]:
        return
        yield

    def __or__[V](self, fn: Callable[[T], V]) -> Option[V]:
        return Nothing()

    def is_some(self) -> bool:
        return False

    def is_some_and(self, p: Callable[[T], bool]) -> bool:
        return False

    def is_nothing(self) -> bool:
        return True

    def is_nothing_or(self, p: Callable[[T], bool]) -> bool:
        return True

    def and_[V](self, o: Option[V]) -> Option[V]:
        return Nothing()

    def and_then[V](self, f: Callable[[], Option[V]]) -> Option[V]:
        return Nothing()

    def or_(self, o: Option[T]) -> Option[T]:
        return o

    def or_else(self, f: Callable[[], Option[T]]) -> Option[T]:
        return f()

    def or_none(self) -> T | None:
        return None

    def xor(self, o: Option[T]) -> Option[T]:
        return o

    def filter(self, p: Callable[[T], bool]) -> Option[T]:
        return Nothing()

    def inspect(self, fn: Callable[[T], None]) -> Option[T]:
        return self

    def unwrap(self) -> Never:
        raise Panic()

    def unwrap_or(self, t: T) -> T:
        return t

    def unwrap_or_else(self, f: Callable[[], T]) -> T:
        return f()

    def expect(self, m: str) -> Never:
        raise Panic(m)

    def map[V](self, fn: Callable[[T], V]) -> Option[V]:
        return Nothing()

    def map_or[V](self, v: V, fn: Callable[[T], V]) -> V:
        return v

    def map_or_else[V](self, v: Callable[[], V], fn: Callable[[T], V]) -> V:
        return v()

    def ok_or[E: Exception](self, e: E) -> Result[T, E]:
        from encrustable.result import Err

        return Err(e)

    def ok_or_else[E: Exception](self, f: Callable[[], E]) -> Result[T, E]:
        from encrustable.result import Err

        return Err(f())

    def zip[V](self, o: Option[V]) -> Option[tuple[T, V]]:
        return Nothing()


type Option[T] = Some[T] | Nothing[T]
